/*
 *  ========================================================================
 *  DIScrete event baSed Energy Consumption simulaTor 
 *    					             for Clouds and Federations (DISSECT-CF)
 *  ========================================================================
 *  
 *  This file is part of DISSECT-CF.
 *  
 *  DISSECT-CF is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *  
 *  DISSECT-CF is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 *  General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with DISSECT-CF.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  (C) Copyright 2013, Gabor Kecskemeti (gkecskem@dps.uibk.ac.at,
 *   									  kecskemeti.gabor@sztaki.mta.hu)
 */


package hu.mta.sztaki.lpds.cloud.simulator.iaas;

import hu.mta.sztaki.lpds.cloud.simulator.Timed;
import hu.mta.sztaki.lpds.cloud.simulator.iaas.VirtualMachine.State;
import hu.mta.sztaki.lpds.cloud.simulator.storage.Repository;
import hu.mta.sztaki.lpds.cloud.simulator.storage.StorageObject;
import hu.mta.sztaki.lpds.cloud.simulator.storage.VirtualAppliance;
import hu.mta.sztaki.lpds.cloud.simulator.util.FastArray;

import java.util.EnumSet;
import java.util.List;

public class PhysicalMachine extends Timed implements
		VMManager<PhysicalMachine> {

	public static enum State {
		OFF,
		SWITCHINGON,
		RUNNING,
		SWITCHINGOFF
	};

	public static enum VMHandlingResult {
		NOPROBLEM, NOTRUNNING, NOTENOUGHRESOURCES, STILLRUNNING
	};

	public static EnumSet<VMHandlingResult> errenousResult = EnumSet
			.complementOf(EnumSet.of(VMHandlingResult.NOPROBLEM));

	private int cpuCoreNum;
	private int promisedCores = 0;
	private long totalMemory;
	private int promisedMemory = 0;
	private final Repository localDisk;

	public final int idleCons;
	public final int maxCons;
	public final int consDiff;
	private double totalConsumption = 0;

	public final int onDelay;
	public final int offDelay;
	private int inDelay = 0;
	private State currentState = State.OFF;

	private final FastArray<VirtualMachine> vms;
	private long completedVMs = 0;

	public PhysicalMachine(int cores, long memory, long disk, long incBW,
			long outBW, long diskBW, int idle, int max, int onD, int offD) {
		cpuCoreNum = cores;
		totalMemory = memory;
		localDisk = new Repository(disk, incBW, outBW, diskBW);

		idleCons = idle;
		maxCons = max;
		consDiff = maxCons - idleCons;

		onDelay = onD;
		offDelay = offD;

		vms = new FastArray<VirtualMachine>(new VirtualMachine[cpuCoreNum + 1]);
	}

	public boolean switchoff() {
		if (vms.size() > 0) {
			return false;
		}
		switch (currentState) {
		case RUNNING:
		case SWITCHINGON:
			inDelay += offDelay;
			currentState = State.SWITCHINGOFF;
		case OFF:
		case SWITCHINGOFF:
		}
		return true;
	}

	public boolean isRunning() {
		return currentState.equals(State.RUNNING);
	}

	public void turnon() {
		switch (currentState) {
		case OFF:
		case SWITCHINGOFF:
			inDelay += onDelay;
			currentState = State.SWITCHINGON;
		case RUNNING:
		case SWITCHINGON:
		}
	}

	public double getTotalConsumption() {
		return totalConsumption;
	}

	public double getLoad() {
		final double[] load = new double[1];
		load[0] = 0;
		vms.iterate(new FastArray.ForEachAction<VirtualMachine>() {
			@Override
			public void eval(VirtualMachine element) {
				load[0] += element.getLoad();
			}
		});
		return load[0] / cpuCoreNum;
	}

	public Repository getLocalDisk() {
		return localDisk;
	}

	@Override
	public void migrateVM(final VirtualMachine vm, final PhysicalMachine target)
			throws VMManagementException {
		if (vms.remove(vm)) {
			vms.add(vm);
			vm.subscribeStateChange(new VirtualMachine.StateChange() {
				@Override
				public void stateChanged(VirtualMachine.State oldState,
						VirtualMachine.State newState) {
					switch (newState) {
					case RUNNING:
						vms.remove(vm);
						completedVMs++;
						target.vms.add(vm);
					case SUSPENDED_MIG:
					case DESTROYED:
						vm.unsubscribeStateChange(this);
					default:
					}
				}
			});
			vm.migrate(target);
		}

	}

	@Override
	public void reallocateResources(final VirtualMachine vm,
			final VirtualAppliance newresources) {
	}

	@Override
	public VirtualMachine requestVM(final VirtualAppliance va,
			Repository vaSource) throws VMManagementException {
		if (!currentState.equals(State.RUNNING)) {
			throw new VMManagementException("Not running");
		}
		final int requestedCores = va.getRequestedCores();
		final long requestedMem = va.getRequestedMem();
		if ((0 <= cpuCoreNum - promisedCores - requestedCores)
				&& (0 <= totalMemory - promisedMemory - requestedMem)) {
			StorageObject foundLocal = localDisk.lookup(va.id);
			StorageObject foundRemote = vaSource == null ? null : vaSource
					.lookup(va.id);
			if (foundLocal != null || foundRemote != null) {
				VirtualMachine vm = new VirtualMachine(va);
				vm.switchOn(this, vaSource);
				vms.add(vm);
				promisedCores += requestedCores;
				promisedMemory += requestedMem;
				vm.subscribeStateChange(new VirtualMachine.StateChange() {
					@Override
					public void stateChanged(VirtualMachine.State oldState,
							VirtualMachine.State newState) {
						if (newState.equals(VirtualMachine.State.RUNNING)) {
							promisedCores -= requestedCores;
							cpuCoreNum -= requestedCores;
							promisedMemory -= requestedMem;
							totalMemory -= requestedMem;
						}
					}
				});
				return vm;
			} else {
				throw new VMManagementException("No VA available!");
			}
		} else {
			return null;
		}
	}

	@Override
	public void terminateVM(final VirtualMachine vm) throws NoSuchVMException,
			VMManagementException {
		if (!vms.remove(vm)) {
			throw new NoSuchVMException(
					"Termination request was received for an unknown VM");
		}
		vm.switchoff();
		completedVMs++;
		cpuCoreNum += vm.getVa().getRequestedCores();
		totalMemory += vm.getVa().getRequestedMem();
	}

	@Override
	public List<VirtualMachine> listVMs() {
		return vms.asList();
	}

	@Override
	public void eval() {
		totalConsumption += currentState.equals(State.OFF) ? 0
				: ((1.0 * idleCons + (consDiff * getLoad())) / 3600);
		if (inDelay > 0) {
			inDelay--;
			if (inDelay == 0) {
				switch (currentState) {
				case SWITCHINGON:
					currentState = State.RUNNING;
					break;
				case SWITCHINGOFF:
					currentState = State.OFF;
					break;
				case RUNNING:
				case OFF:
				}
			}
		} else if (currentState.equals(State.RUNNING)) {
		}
	}
}
