/*
 *  ========================================================================
 *  DIScrete event baSed Energy Consumption simulaTor 
 *    					             for Clouds and Federations (DISSECT-CF)
 *  ========================================================================
 *  
 *  This file is part of DISSECT-CF.
 *  
 *  DISSECT-CF is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *  
 *  DISSECT-CF is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 *  General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with DISSECT-CF.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  (C) Copyright 2013, Gabor Kecskemeti (gkecskem@dps.uibk.ac.at,
 *   									  kecskemeti.gabor@sztaki.mta.hu)
 */


package hu.mta.sztaki.lpds.cloud.simulator.iaas;

import hu.mta.sztaki.lpds.cloud.simulator.Timed;
import hu.mta.sztaki.lpds.cloud.simulator.iaas.VMManager.VMManagementException;
import hu.mta.sztaki.lpds.cloud.simulator.storage.NetworkNode;
import hu.mta.sztaki.lpds.cloud.simulator.storage.Repository;
import hu.mta.sztaki.lpds.cloud.simulator.storage.StorageObject;
import hu.mta.sztaki.lpds.cloud.simulator.storage.VirtualAppliance;
import hu.mta.sztaki.lpds.cloud.simulator.util.FastArray;

import java.util.EnumSet;

public class VirtualMachine extends Timed {

	public static class StateChangeException extends VMManagementException {
		private static final long serialVersionUID = 2950595344006507672L;

		public StateChangeException(final String e) {
			super(e);
		}

	}

	public static interface StateChange {
		void stateChanged(State oldState, State newState);
	}

	public static interface LoadProvider {
		public float getCurrentload();
	}

	private interface EventSetup {
		public void changeEvents();
	}

	private VirtualAppliance va;
	private StorageObject disk = null;
	private StorageObject savedmemory = null;
	private PhysicalMachine host = null;
	private Repository vasource = null;

	public static enum State {
		INITIAL_TR,
		STARTUP,
		RUNNING,
		SUSPEND_TR,
		SUSPENDED,
		SUSPENDED_MIG,
		RESUME_TR,
		MIGRATING,
		SHUTDOWN,
		DESTROYED
	};

	public final static EnumSet<State> consumingStates = EnumSet.of(
			State.STARTUP, State.RUNNING, State.MIGRATING, State.RESUME_TR);
	public final static EnumSet<State> transferringStates = EnumSet.of(
			State.INITIAL_TR, State.SUSPEND_TR, State.RESUME_TR,
			State.MIGRATING);
	public final static EnumSet<State> suspendedStates = EnumSet.of(
			State.SUSPENDED, State.SUSPENDED_MIG);
	public final static EnumSet<State> preStartupStates = EnumSet.of(
			State.DESTROYED, State.SHUTDOWN);

	private FastArray<StateChange> subscribers = new FastArray<StateChange>(
			new StateChange[5]);

	private State currState = State.DESTROYED;

	public static final float loadwhilenotrunning = 0.2f;

	private LoadProvider myload = new LoadProvider() {
		@Override
		public float getCurrentload() {
			return 1;
		}
	};

	public VirtualMachine(final VirtualAppliance va) {
		this.va = va;
	}

	private void setState(final State newstate) {
		if (subscribers.size() > 0) {
			subscribers
					.iterate(new FastArray.ForEachAction<VirtualMachine.StateChange>() {
						@Override
						public void eval(StateChange element) {
							element.stateChanged(currState, newstate);
						}
					});
		}
		currState = newstate;
	}

	public VirtualAppliance getVa() {
		return va;
	}

	public State getState() {
		return currState;
	}

	public void setLoadProvider(final LoadProvider lp) {
		if (lp != null) {
			myload = lp;
		}
	}

	public float getLoad() {
		return consumingStates.contains(currState) ? (currState == State.RUNNING ? myload
				.getCurrentload() : loadwhilenotrunning)
				: 0;
	}

	public void prepare(final PhysicalMachine pm, final Repository vasource)
			throws StateChangeException, VMManagementException {
		if (currState != State.DESTROYED) {
			throw new StateChangeException("The VM is not destroyed");
		}
		initialTransfer(pm, vasource, new EventSetup() {
			@Override
			public void changeEvents() {
				setState(State.SHUTDOWN);
			}
		});
	}

	private void initialTransfer(final PhysicalMachine pm,
			final Repository vasource, final EventSetup es)
			throws VMManagementException {
		final State oldState = currState;
		setState(State.INITIAL_TR);
		host = pm;
		this.vasource = vasource;
		final String diskid = Integer.toString((hashCode()));
		final long bgnwload = va.getBgNetworkLoad();
		class InitialTransferEvent implements NetworkNode.NetworkEvent {
			final Repository target;

			public InitialTransferEvent(Repository t) {
				target = t;
			}

			@Override
			public void trComplete() {
				disk = target.lookup(diskid);
				es.changeEvents();
			}
		}
		boolean noerror;
		Repository pmdisk = pm.getLocalDisk();
		if (bgnwload > 0) {
			noerror = vasource.duplicateContent(va.id, diskid,
					new InitialTransferEvent(vasource));
		} else {
			if (vasource == null) {
				noerror = pmdisk.duplicateContent(va.id, diskid,
						new InitialTransferEvent(pmdisk));
			} else {
				noerror = vasource.requestContentDelivery(va.id, diskid,
						pmdisk, new InitialTransferEvent(pmdisk));
			}
		}
		if (!noerror) {
			setState(oldState);
			throw new VMManagementException("Initial transfer failed");
		}
	}

	public void switchOn(final PhysicalMachine pm, final Repository vasource)
			throws StateChangeException, VMManagementException {
		if (!preStartupStates.contains(currState)) {
			throw new StateChangeException(
					"The VM is not shut down or destroyed");
		}
		EventSetup ev = new EventSetup() {
			@Override
			public void changeEvents() {
				setState(State.STARTUP);
				startBGLoad();
			}
		};
		if (currState == State.DESTROYED) {
			initialTransfer(pm, vasource, ev);
		} else {
			ev.changeEvents();
		}
	}

	private void resumeAfterMigration(final PhysicalMachine pm) {
		final PhysicalMachine oldHost = host;
		host = pm;
		try {
			resume();
			final Repository previousDisk = oldHost.getLocalDisk();
			previousDisk.deregisterObject(disk);
			previousDisk.deregisterObject(savedmemory);
		} catch (StateChangeException e) {
			System.err.println("IMPROPER STATE DURING MIGRATION!");
		} catch (VMManagementException e) {
			host = oldHost;
			setState(State.SUSPENDED_MIG);
		}
	}

	private void actualMigration(final PhysicalMachine pm) {
		final boolean[] cancelMigration = new boolean[1];
		cancelMigration[0] = false;
		class MigrationEvent implements NetworkNode.NetworkEvent {
			int eventcounter = 0;

			@Override
			public void trComplete() {
				if (cancelMigration[0]) {
					pm.getLocalDisk().deregisterObject(disk);
				} else {
					eventcounter++;
					if (eventcounter == 2) {
						resumeAfterMigration(pm);
					}
				}
			}
		}

		if (va.getBgNetworkLoad() > 0) {
			resumeAfterMigration(pm);
		} else {
			final Repository from = host.getLocalDisk();
			final Repository to = pm.getLocalDisk();
			final MigrationEvent mp = new MigrationEvent();
			if (from.requestContentDelivery(disk.id, to, mp)) {
				if (from.requestContentDelivery(savedmemory.id, to, mp)) {
					return;
				}
				cancelMigration[0] = true;
			}
			setState(State.SUSPENDED_MIG);
		}
	}

	public void migrate(final PhysicalMachine pm) throws StateChangeException,
			VMManagementException {
		if (suspendedStates.contains(currState)) {
			actualMigration(pm);
		} else {
			suspend(new EventSetup() {
				@Override
				public void changeEvents() {
					setState(State.SUSPENDED);
					actualMigration(pm);
				}
			});
		}
	}

	public void destroy() throws StateChangeException {
		if (transferringStates.contains(currState)) {
			throw new StateChangeException(
					"Parts of the VM are under transfer."
							+ "This transfer should be finished before destruction.");
		}
		if (host != null) {
			host.getLocalDisk().deregisterObject(disk);
			host.getLocalDisk().deregisterObject(savedmemory);
			host = null;
		}
		if (vasource != null) {
			vasource.deregisterObject(disk);
			vasource.deregisterObject(savedmemory);
			vasource = null;
		}
		setState(State.DESTROYED);
	}

	public void switchoff() throws StateChangeException {
		if (currState != State.RUNNING) {
			throw new StateChangeException(
					"Cannot switch off a not running machine");
		}
		setState(State.SHUTDOWN);
	}

	public void suspend() throws StateChangeException, VMManagementException {
		suspend(new EventSetup() {
			@Override
			public void changeEvents() {
				setState(State.SUSPENDED);
			}
		});
	}

	private void suspend(final EventSetup ev) throws StateChangeException,
			VMManagementException {
		if (currState != State.RUNNING) {
			throw new StateChangeException(
					"Cannot suspend a not running machine");
		}
		final String memid = "VM-Memory-State" + hashCode();
		final String tmemid = "Temp-" + memid;
		final Repository pmdisk = host.getLocalDisk();
		savedmemory = new StorageObject(tmemid, va.getRequestedMem());
		if (pmdisk.registerObject(savedmemory)) {
			throw new VMManagementException(
					"Not enough space on localDisk for the suspend operation of "
							+ hashCode());
		}
		setState(State.SUSPEND_TR);
		final long bgnwload = va.getBgNetworkLoad();
		class SuspendComplete implements NetworkNode.NetworkEvent {
			@Override
			public void trComplete() {
				pmdisk.deregisterObject(savedmemory);
				savedmemory = bgnwload > 0 ? vasource.lookup(memid) : pmdisk
						.lookup(memid);
				ev.changeEvents();
			}
		}
		;
		boolean trsuccess;
		if (bgnwload > 0) {
			trsuccess = pmdisk.requestContentDelivery(tmemid, memid, vasource,
					new SuspendComplete());
		} else {
			trsuccess = pmdisk.duplicateContent(tmemid, memid,
					new SuspendComplete());
		}
		if (!trsuccess) {
			setState(State.RUNNING);
			throw new VMManagementException(
					"Not enough space on localDisk for the suspend operation of "
							+ hashCode());
		}
	}

	public void resume() throws StateChangeException, VMManagementException {
		if (!suspendedStates.contains(currState)) {
			throw new StateChangeException(
					"Cannot resume a not suspended machine");
		}
		setState(State.RESUME_TR);
		final String tmemid = "Temp-" + savedmemory.id;
		final Repository pmdisk = host.getLocalDisk();
		final long bgnwload = va.getBgNetworkLoad();
		class ResumeComplete implements NetworkNode.NetworkEvent {
			@Override
			public void trComplete() {
				pmdisk.deregisterObject(tmemid);
				if (bgnwload > 0) {
					vasource.deregisterObject(savedmemory);
				} else {
					pmdisk.deregisterObject(savedmemory);
				}

				savedmemory = null;
				setState(State.RUNNING);
			}
		}
		;
		boolean trsuccess;
		if (bgnwload > 0) {
			startBGLoad();
			trsuccess = vasource.requestContentDelivery(savedmemory.id, tmemid,
					pmdisk, new ResumeComplete());
		} else {
			trsuccess = pmdisk.duplicateContent(savedmemory.id, tmemid,
					new ResumeComplete());
		}
		if (!trsuccess) {
			setState(State.SUSPENDED);
			throw new VMManagementException(
					"Not enough space on localDisk for the resume operation of "
							+ hashCode());
		}
	}

	public void subscribeStateChange(final StateChange consumer) {
		subscribers.add(consumer);
	}

	public void unsubscribeStateChange(final StateChange consumer) {
		subscribers.remove(consumer);
	}

	private class BGLoadEvent implements NetworkNode.NetworkEvent {
		@Override
		public void trComplete() {
			if (consumingStates.contains(currState)) {
				startBGLoad();
			}
		}
	}

	private BGLoadEvent loadEvent = new BGLoadEvent();

	private void startBGLoad() {
		final long bgnwload = va.getBgNetworkLoad();
		if (bgnwload > 0) {
			NetworkNode.initTransfer(bgnwload, host.getLocalDisk(), vasource,
					loadEvent);
		}
	}

	public void changeVA(final VirtualAppliance newVA,
			final Repository newVASource) throws VMManagementException {
		switch (currState) {
		case INITIAL_TR:
		case RESUME_TR:
		case SUSPEND_TR:
		case MIGRATING:
		case SUSPENDED:
		case SUSPENDED_MIG:
			throw new StateChangeException(
					"Cannot change VA while the VM is in a transferring or stored state");
		case RUNNING:
		case STARTUP:
			if (newVA == va && newVASource == vasource) {
				va = newVA;
			} else {
				if (vasource == null || va.getBgNetworkLoad() <= 0) {
					host.getLocalDisk().deregisterObject(disk);
				} else {
					vasource.deregisterObject(disk);
				}
				va = newVA;
				setState(State.DESTROYED);
				switchOn(host, newVASource);
			}
			break;
		case SHUTDOWN:
			if (vasource == null || va.getBgNetworkLoad() <= 0) {
				host.getLocalDisk().deregisterObject(disk);
			} else {
				vasource.deregisterObject(disk);
			}
			va = newVA;
			setState(State.DESTROYED);
			prepare(host, newVASource);
			break;
		case DESTROYED:
			va = newVA;
			vasource = null;
		}
	}

	@Override
	public void eval() {
	}
}
