/*
 *  ========================================================================
 *  DIScrete event baSed Energy Consumption simulaTor 
 *    					             for Clouds and Federations (DISSECT-CF)
 *  ========================================================================
 *  
 *  This file is part of DISSECT-CF.
 *  
 *  DISSECT-CF is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *  
 *  DISSECT-CF is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 *  General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with DISSECT-CF.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  (C) Copyright 2013, Gabor Kecskemeti (gkecskem@dps.uibk.ac.at,
 *   									  kecskemeti.gabor@sztaki.mta.hu)
 */


package hu.mta.sztaki.lpds.cloud.simulator.storage;

import hu.mta.sztaki.lpds.cloud.simulator.Timed;
import hu.mta.sztaki.lpds.cloud.simulator.util.FastArray;

public class NetworkNode {

	public static interface NetworkEvent {
		public void trComplete();
	}

	static class SingleTransfer {
		public long currentContent;
		public long totalTransfer;
		final public NetworkEvent ev;

		final public NetworkNode target;

		private SingleTransfer(final long tottr, final NetworkEvent e,
				NetworkNode to) {
			currentContent = 0;
			totalTransfer = tottr;
			ev = e;
			target = to;
		}
	}

	private abstract class BandwidthSpreader extends Timed {
		protected final long resetBW;
		protected final FastArray<SingleTransfer> tg = new FastArray<SingleTransfer>(
				new SingleTransfer[100]);
		protected int tgsize = 0;

		public BandwidthSpreader(final long initialBW) {
			resetBW = initialBW;
		}

		public void registerTR(final SingleTransfer st) {
			tg.add(st);
			tgsize++;
			if (tgsize == 1) {
				subscribe();
			}
		}

		public void deregisterTR(final SingleTransfer st, final int i) {
			st.ev.trComplete();

			tg.remove(i);
			tgsize--;
			if (tgsize == 0) {
				unsubscribe();
			}

			final BandwidthSpreader sourcebws;
			if (st.target.diskinbws == this) {
				sourcebws = st.target.diskoutbws;
			} else {
				sourcebws = st.target.outbws;
			}
			sourcebws.tg.remove(st);
			sourcebws.tgsize--;
			if (sourcebws.tgsize == 0) {
				sourcebws.unsubscribe();
			}
		}

		@Override
		public void eval() {
			long remainingBandwidth = resetBW;
			long pastRemaining;
			try {
				do {
					pastRemaining = remainingBandwidth;
					remainingBandwidth = singleSpread(tg.getManagedArray(), pastRemaining / tgsize);

				} while (remainingBandwidth > 0
						&& pastRemaining != remainingBandwidth);
			} catch (ArithmeticException e) {
			}
		}

		protected abstract long singleSpread(final SingleTransfer[] arr, final long toSpread);
	}

	private class InputSpreader extends BandwidthSpreader {

		public InputSpreader(final long initialBW) {
			super(initialBW);
		}

		@Override
		protected long singleSpread(final SingleTransfer[] arr, final long toSpread) {
			long remainingBandwidth = 0;
			for (int i = tgsize - 1; 0 <= i; i--) {
				final SingleTransfer st = arr[i];
				final long reducedContent = st.currentContent - toSpread;
				if (reducedContent > 0) {
					st.currentContent = reducedContent;
				} else {
					remainingBandwidth -= reducedContent;
					if (st.totalTransfer == 0) {
						deregisterTR(st, i);
					} else {
						st.currentContent = 0;
					}
				}
			}
			return remainingBandwidth;
		}
	}

	private class OutputSpreader extends BandwidthSpreader {
		public OutputSpreader(final long initialBW) {
			super(initialBW);
		}

		@Override
		protected long singleSpread(final SingleTransfer[] arr, final long toSpread) {
			long remainingBandwidth = toSpread * tgsize;
			for (int i = tgsize - 1; 0 <= i; i--) {
				final SingleTransfer st = arr[i];
				final long realRequestedContent = st.currentContent
						+ (toSpread < st.totalTransfer ? toSpread
								: st.totalTransfer);
				final long cc = realRequestedContent < resetBW ? realRequestedContent
						: resetBW;
				final long transferdiff = cc - st.currentContent;
				st.currentContent = cc;
				st.totalTransfer -= transferdiff;

				remainingBandwidth -= transferdiff;
			}
			return remainingBandwidth;
		}
	}


	private final BandwidthSpreader inbws;
	private final BandwidthSpreader outbws;
	private final BandwidthSpreader diskinbws;
	private final BandwidthSpreader diskoutbws;

	public NetworkNode(final long maxInBW, final long maxOutBW,
			final long diskBW) {
		outbws = new OutputSpreader(maxOutBW);
		inbws = new InputSpreader(maxInBW);
		diskinbws = new InputSpreader(diskBW / 2);
		diskoutbws = new OutputSpreader(diskBW / 2);
	}

	public long getOutputbw() {
		return outbws.resetBW;
	}

	public long getInputbw() {
		return inbws.resetBW;
	}

	public long getDiskbw() {
		return diskinbws.resetBW * 2;
	}

	public static void initTransfer(final long size, final NetworkNode from,
			final NetworkNode to, final NetworkEvent e) {
		final SingleTransfer st = new SingleTransfer(size, e, to);
		if (from == to) {
			from.diskinbws.registerTR(st);
			from.diskoutbws.registerTR(st);
		} else {
			from.inbws.registerTR(st);
			to.outbws.registerTR(st);
		}
	}
}
